unit CompilU1;


interface


const
  ProgramName   = '6502 Pascal Compiler';

  conTrue       =    $FF;           { True }
  conFalse      =    $00;           { False }

  BooleanSize   =      1;
  ByteSize      =      1;
  CharSize      =      1;
  IntSize       =      2;
  LabelSize     =      0;
  LongSize      =      4;
  PathFile      =     40;           { max. length drive + path + filename }
  PointerSize   =      2;           { for the moment 16-bits }
  RealSize      =      6;
  SetItemSize   =      1;
  StringSize    =    255;
  WordSize      =      2;
  MaxDigits     =     21;           { max. length when storing number as text }
  MaxLngIdent   =     19;           { maximum length of identifiers }
  MaxSubArray   =      4;           { max. ranges declared within one array }
  MaxTop        =     20;           { max. level of display }
  MaxProcParms  =     20;           { max. level of nested procedures }
  MaxRange      =      4;           { max. size of range terms
                                       2 = byte, 4 = integer, 8 = longint }
  MaxUnits      =     32;           { max. number of units }
  MaxWith       =     20;           { max. level of WITH structure }



type
  alfa    = string[MaxLngIdent];
  Str3    = string[3];
  Str16   = string[16];
  Str80   = string[80];


  tSymbol = (
    ByteVal, CharVal, IntVal, LongVal, RangeVal, RealVal, StringVal, WordVal,

    ident, plus, minus, star, lbrack, rbrack, colon, eql, neq, lss, leq, gtr,
    geq, lParent, rParent, comma, semicolon, period, becomes, unknownsym,
    rangesym, slash, StanFunc, dakje, SetSym, SetItemSym, To_do, AtSym,

    AllSym, andsym, arraysym, AsmSym, beginsym, BoolSym, bytesym, casesym,
    charsym, CommentSym, CompDir1Sym, CompDir2Sym, constsym, divsym, dosym,
    downtosym, elsesym, endsym, ExitSym, FalseSym, filesym, forsym,
    forwardsym, funcsym, gotosym, ifsym, ImplemSym, InSym, intsym, labelsym,
    longsym, MemSym, MemWSym, MemLSym, modsym, NilSym, notsym, packedsym,
    pointsym, PortSym, PortWSym, procsym, progsym, ofsym, orsym,
    realsym, recordsym, repeatsym, SetOfSym, ShlSym,
    ShrSym, stringsym, TextSym, ThenSym, tosym, TrueSym, typesym, UnitSym,
    untilsym, UsesSym, varsym, whilesym, withsym, wordsym, WriteSym,
    WritelnSym, XorSym,

    AbsSym, ChrSym, FileSizSym, HiSym, LengthSym, LoSym, OddSym, OrdSym,
    PredSym, RoundSym, SizeOfSym, SuccSym, SwapSym, TruncSym
   );

  SetOfSymbol = set of tSymbol;

  tGroup = (types, konst, vars, ProcFunc, FPforward, unknown, ASMstring);

  tRangeType = (rChar, rInt, rSet);

  pString = ^string;

  pCaseStruc  = ^tCaseStruc;
  pFileDef    = ^tFileDef;
  pIdentifier = ^tIdentifier;
  pRange      = ^tRange;
  pRealArray  = ^tRealArray;
  pStructure  = ^tStructure;
  pWithStruc  = ^tWithStruc;

  Result = Record
    Name : alfa;
    OrgName : alfa;
    Offset : longint;                   { if Vars: offset in case of array }
                                        { if ASM : start address of result }
    Symbol : tSymbol;
    Group : tGroup;                     { constant, variable or ASM-strings }
    boDirect,                           { direct or indirect address }
    b : boolean;                        { store value if not variable }
    c : char;
    l : longint;                        { size of item if not variable }
    r : real;
    s : pString;
    StrSize : word;
    Range : pRange;
    Ident : pIdentifier;                { store pointer of found variable }
   end;

  tCaseStruc = record
    CsubPart : pIdentifier;
    NextCase : pCaseStruc;
    case CsubType : tSymbol of
     ByteVal,
     WordVal,
     IntVal,
     LongVal : (l : longint);
     CharVal : (c : Char);
     StringVal : (s : pString);
     Ident : (CsubName : alfa);
   end;

  tRange = record
    nRange    : pRange;
    nRaOrg    : boolean;                { original nRange or copy }
    RangeType : tRangeType;
    lBegin, lEnd : longint;
    SetId : pIdentifier;                { origin of set }
   end;

  tRealArray = array[0..9] of byte;
  tArFFFE = array[0..$FFFE] of byte;
  pArFFFE = ^tArFFFE;

  tFileDef = record
    Filename  : string[PathFile];
    fType     : tSymbol;                { Type, byte, char etc. }
    fTypeId   : pIdentifier;            { if Type then Identifier }
    BlockSize : word;
    UEX       : Str3;                   { user supplied extension }
    Attribute : Byte;
{
  bit 7:  1 = file is closed,  0 = not closed
  bit 6:  1 = file is locked,  0 = not locked

  bit 5 4
      0 0  reset
      0 1  rewrite
      1 0  append

  bit 3..0  = filetype
         0  = DELeted
         1  = SEQuential
         2  = PRoGram
         3  = USeR
         4  = RELative
         5  = DIRectory (1581 compatibility, not used)
         6  = DIRectory
         7  = SYStem
         8  = UEX, User supplied EXtension
         9  = EXE
         A  = 1541-image
         B  = 1571-image
         C  = 1581-image
         D  = 8050-image
         E  = 8250-image
         F  = D16-image / CMD-partition
}
   end;

  tStructure = record
    boPointer : boolean;                { type is like '^array ....' }
    case sSymbol : tSymbol of
     ArraySym : (
       ArRange  : pRange;
       ArOrg    : boolean;              { original ArRange or copy }
       ArNum    : byte;                 { number of subranges }
       ArSize,                          { size of item }
       DataSize : longint;              { size of array }
       DataPtr  : pArFFFE;
       ArStruc  : pStructure);
     BoolSym : (b : boolean);
     ByteSym,
     IntSym,
     LongSym,
     WordSym  : (l : longint);
     CharSym : (c : char);
     TextSym,
     FileSym : (f : pFileDef);
     LabelSym : ();
     PointSym : ();
     FuncSym,
     ProcSym : (
       IsCompiled : boolean;
       TotSize : word;                  { size of all used varaibles }
       Parameters : pIdentifier;
       Functie : pStructure;
       OrgDeclaration : pString);
     RangeSym : (Range : pRange);
     RealSym : (
       r : real;
       rs : pString);
     RecordSym : (
       Item : pIdentifier;
       CaseItem : pIdentifier;
       CaseMembers : pIdentifier;
       CaseStruc : pCaseStruc);
     SetSym : (tSet : pIdentifier);
     SetItemSym : (si : byte);
     SetOfSym : (tSetOf : pIdentifier);
     StringSym : (
       StrSize : word;
       s : pString);
     TypeSym : (t : pIdentifier);
   end;

  tIdentifier = record
    Name,
    AssName,
    UnitName : alfa;
    Group : tGroup;
    Level : byte;                       { 1 = stored in INTERFACE
                                          0 = stored in IMPLEMENTATION }
    boUsed,                             { is it used some where? }
    ProcVar : boolean;
    Size,                               { size of variable }
    Start : longint;                    { startaddress where data is stored
                                           or place of item in set }
    nlink : pIdentifier;
    Structure : pStructure;
   end;

  tPlaceInMem = record
    bType : byte;                       { 0 = before program
                                          1 = after program/variables
                                          2 = use address   }
    Address : longint;
   end;

  pUnitList = ^tUnitList;
  tUnitList = record
    UnitName : alfa;
    NextUnit : pUnitList
   end;

  tWithStruc = record
    ThisWith : pIdentifier;
    NextWith : pWithStruc;
   end;


var
  sFunctionNow,                         { name of function now being processed }
  sPasFile,
  id            : alfa;                 { found string in GetSymbol }

  Keyword       : array[AndSym..XorSym] of alfa;
  aFunction     : array[AbsSym..TruncSym] of alfa;
  idProcParms   : array [0..MaxProcParms] of pIdentifier;
  Identifier    : array [0..MaxUnits, 0..MaxTop] of pIdentifier;
  WithStruc     : array [1..MaxWith] of pWithStruc;

  ListUnits     : array[0..MaxUnits] of record
    ProgName : alfa;
    UnitList : pUnitList;
   end;

  boAbsolute,
  boASMcode,
  boComment,
  boHeapUsed,
  boReported,
  boStored00,                           { stored constant $00 }
  boStoredFF,                           { stored constant $FF }
  EndOfLine,                            { end-of-line detected }
  EndOfProg,                            { Variables start at end of program }
  UpdateStart                           { update lStart in CalcStart }
                : boolean;

  BlockLevel,
  bNowUnit,                             { unit now in focus }
  chCounter,                            { number of characters in line }
  NrErrors,                             { number of errors in one line }
  Top,                                  { level of var. Identifier }
  UnitLevel,
  WithTop                               { level of WITH statements }
                : byte;

  ch                                    { last read character }
                : char;

  lBOrgAddress,                         { start of BASIC }
  lFound,                               { last found byte/integer/word/longint }
  lStack,                               { Negative offset to OrgStack,
                                          counts down

                                          !! Points to last occupied place !!
                                        }
  lStart,                               { keeps record of startaddress of
                                          variables and constants, counts up,
                                          points to first free space }
  lStartAddr,                           { start of ML }
  OrgDStack     : longint;              { Dummy stack as 6502 stack is only
                                          256 bytes }

  rFound                                { last found real }
                : real;

  tReM          : Result;

  reAbsolute    : record
    AbsH, AbsL : longint;
    AbsN : Alfa;
   end;

  sBreak,
  sContinue,
  sCPU,
  sExit,
  sMachine      : Str16;

  PasCode,
  sRegel,
  sString                               { contains found string/number }
                : string;

  ASMbody,
  ASMeql,
  ASMfil,
  ASMheader,
  ASMmid,
  fInput        : text;

  PrevSymbol,                           { previous symbol }
  Symbol,                               { last found symbol }
  fSymbol                               { last found standardfunction }
                : tSymbol;

  PlaceHeap,
  PlaceVar      : tPlaceInMem;

  LabelCounter,
  LineCount,
  wError,
  wKonCounter,
  wVarCounter   : word;

  AllCharSys,
  AllNumberSys,
  BlockSys,
  ConstSys,
  NumberSys     : SetOfSymbol;


  tIdRB : pIdentifier;


{============================================================================}


function  Array2Symbol(tSt: pStructure) : tSymbol;
function  ASMline(Re1 : string) : string;
function  CheckPointer(tId : pIdentifier) : boolean;
function  CheckSymbol(sy : tSymbol) : boolean;
function  CompareSymbol(Sy1, Sy2 : tSymbol; Op : tSymbol) : boolean;
function  GetCheckSymbol(sy : tSymbol) : boolean;
function  InsertProgram : boolean;
function  KonName : alfa;
function  NewIdent : pIdentifier;
function  NextLabel : Str16;
function  NumberSymbol(l : longint) : tSymbol;
function  Real2String(r : real) : Str80;
function  Search4Set(id : alfa) : pIdentifier;
Function  SearchChainId(tId : pIdentifier; Naam : Alfa;
                         var Start : longint) : pIdentifier;
function  SearchId(id, UNaam : alfa; Level : byte;
                      boUsed : boolean) : pIdentifier;
function  SearchIdAll(id, UNaam : alfa; Level : byte;
                      boUsed : boolean) : pIdentifier;
function  SearchInSet(id, UNaam : alfa; tId : pIdentifier) : boolean;
function  SearchSet(Name : alfa; tIdS : pIdentifier) : longint;
function  Subsize(tSt : pStructure; bo : boolean) : longint;
function  Symbol2char(s : tSymbol) : str3;
function  Symbol2Num(s : tSymbol) : byte;
function  Symbol2Size(s : tSymbol) : word;
function  Symbol2Type(s : tSymbol) : Str3;
function  tReSize(tRe : Result) : word;
function  Val2Sym(s : tSymbol) : tSymbol;
function  VarName(tId : pIdentifier) : alfa;
procedure  CalcResult(tSt, tSt2 : pStructure; Op1 : tSymbol);
procedure  CalcStart(tId : pIdentifier);
procedure  CalcTerm(tSt : pStructure);
procedure  Calculation(tSt : pStructure; level : byte);
procedure  ClearTre(var tRe : Result);
procedure  Compiler1Directive;
procedure  Compiler2Directive;
procedure  CopyASMheader(stFilename : str80);
procedure  CopyStructure(tSt, tSt2 : pStructure);
procedure  CreateTextStruc(tSt : pStructure);
procedure  DisposeIdentifier(var tIds : pIdentifier);
procedure  DisposeRange(var tRa1 : pRange);
procedure  DisposeStructure(var tSts : pStructure);
procedure  EnterIdGlobal(tIdent: pIdentifier);
procedure  EnterIdLocal(tIdent : pIdentifier; var tIdD : pIdentifier);
procedure  Error(ErrNr : integer);
procedure  GetChar;
procedure  GetFile(tSt : pStructure);
procedure  GetSymbol;
procedure  HandleAbsolute;
procedure  InsertUnit(PName : alfa);
procedure  SkipDirective;
procedure  Struc2Result(var tRe1 : Result; tSt : pStructure; boPnt : boolean);
procedure  Wr2ASMbodyLn(St : string);
procedure  Wr2ASMheader(St : string);
procedure  Wr2ASMheaderLn(St : string);
procedure  Wr2ASMmidLn(St : string);
procedure  WriteConst;
procedure  WriteVars;


{============================================================================}


implementation


uses RB_unit0;


function Array2Symbol(tSt: pStructure) : tSymbol;
var
  Sy : tSymbol;
begin

{#### lot of optimizing to do !!!!! }

  Sy := tSt^.ArStruc^.sSymbol;

  if Sy = TypeSym then
   begin
    tSt := tSt^.ArStruc;
    Sy := tSt^.t^.Structure^.sSymbol;

    if Sy = TypeSym then
     begin
      tSt := tSt^.t^.Structure;

      if tSt^.boPointer then Sy := PointSym
       else Sy := tSt^.t^.Structure^.sSymbol;
     end;
   end;

  Array2Symbol := Sy;
end;


function ASMline(Re1 : string) : string;
var
  by : byte;
  Re2 : str80;
label L1;

 function lengte(Regel : str80) : byte;
 var
   by : byte;
   pst : str80;
 begin
   pst := '';
   by := 1;

   while (by <= length(Regel)) do
    begin
     if (Regel[by] <> #9) then pst := pst + Regel[by] else
      begin
       pst := pst + ' ';
       while (length(pst) mod 8 > 0) do pst := pst + ' ';
      end;

     inc(by);
    end;

   lengte := length(pst);
 end;

begin
  Re2 := '';

  if (Re1 = '') then ASMline := ''
   else

  if (Re1[1] = ';') then ASMline := Re1 else
   begin
    if (Re1[1] in [' ', #9]) then
     begin
      Re1 := short(Re1);
      if (Re1 = '') then goto L1;

      Re2 := #9#9;
     end
      else

    if (Re1[1] <> '.') then
     begin
      by := 0;
      while (not (Re1[by+1] in [' ', #9]) and (by < length(Re1))) do inc(by);
      Re2 := copy(Re1, 1, by);
      Delete(Re1, 1, by);
      Re1 := short(Re1);
      if (Re1 = '') then goto L1;

      while (lengte(Re2) < 16) do Re2 := Re2 + #9;
     end;

    if (Re1[1] = '.') then
     begin
      if (   (upper(copy(Re1, 2, 2)) = 'EQ')
          or (upper(copy(Re1, 2, 2)) = 'AL')) then
       begin
        by := 0;
        while ((Re1[by+1] <> '=') and (by < length(Re1))) do inc(by);
        Re2 := Re2 + copy(Re1, 1, by);
        if (lengte(Re2) < 16) then Re2 := Re2 + #9;
        Re2 := Re2 + '= ';
        Delete(Re1, 1, by+1);
        Re1 := short(Re1);
        if (Re1 = '') then goto L1;
       end;

      by := 0;
      while ((Re1[by+1] <> ';') and (by < length(Re1))) do inc(by);
      Re2 := Re2 + copy(Re1, 1, by);
      Delete(Re1, 1, by);
      Re1 := short(Re1);
     end
    else
     begin
      by := 0;
      while (not (Re1[by+1] in [' ', #9]) and (by < length(Re1))) do inc(by);
      Re2 := Re2 + copy(Re1, 1, by);
      Delete(Re1, 1, by);
      Re1 := short(Re1);
      if (Re1 = '') then goto L1;

      Re2 := Re2 + #9;
     end;

    if (Re1[1] <> ';') then
     begin
      by := 0;
      while ((Re1[by+1] <> ';') and (by < length(Re1))) do inc(by);
      Re2 := Re2 + copy(Re1, 1, by);
      Delete(Re1, 1, by);
      Delete(Re1, 1, by);
      if (Re1 = '') then goto L1;

      Re2 := Re2 + #9;
     end;

    while (lengte(Re2) < 40) do Re2 := Re2 + #9;
    Re2 := Re2 + Re1;

L1:
    ASMline := Re2;
   end;
end;


function CheckPointer(tId : pIdentifier) : boolean;
var
  tSt: pStructure;
label L1, L2;
begin
  CheckPointer := False;
  tSt := tId^.Structure;

L1:
  if tSt^.boPointer then
   begin
       {#### 20040930}
  CheckPointer := True;
  exit;

    if tSt^.sSymbol = TypeSym then
     begin
L2:   tId := tSt^.t;
      if not tId^.Structure^.boPointer then exit;
     end;
   end
  else
   begin
    if tSt^.sSymbol = ArraySym then
     begin
      tSt := tSt^.ArStruc;
      goto L1;
     end
      else
    if tSt^.sSymbol <> TypeSym then exit;
    goto L2;
   end;

  CheckPointer := True;
end;


function CheckSymbol(sy : tSymbol) : boolean;
var
  w : word;
begin
  CheckSymbol := True;

  if (Symbol <> sy) then
   begin
    case sy of
      ImplemSym : w := 73;
      SemiColon : w := 85;
      colon : w := 86;
      comma : w := 87;
      lParent : w := 88;
      rParent : w := 89;
      eql : w := 90;
      becomes : w := 91;
      lBrack : w := 92;
      rBrack : w := 93;
      period : w := 94;
      RangeSym : w := 95;
       else w := 999;
     end;

    Error(w);
    CheckSymbol := False;
   end;
end;


function CompareSymbol(Sy1, Sy2 : tSymbol; Op : tSymbol) : boolean;
{
 Check if the given types and operator match with eachother. In fact only
 the combinations that are NOT allowed are checked.
}
begin
  CompareSymbol := False;

  if (   ((Sy2 = BoolSym) and not(Sy1 = BoolSym))
      or ((Sy1 = BoolSym) and not(Sy2 = BoolSym))) then exit;

  if (Op = plus) then
   begin
    if (   (    (Sy2 in [SetItemSym, SetSym, SetOfSym])
            and not(Sy1 in [SetItemSym, SetSym, SetOfSym]))
        or (    (Sy1 in [SetItemSym, SetSym, SetOfSym])
            and not(Sy2 in [SetItemSym, SetSym, SetOfSym]))) then exit;
   end;

  if (Op = InSym) then
   begin
    if (Sy1 in [SetItemSym, SetSym]) then
     begin
      if not(Sy2 in [SetItemSym, SetSym, SetOfSym]) then exit;
     end
      else

    if (Sy1 in NumberSys) then
     begin
      if not (Sy2 in NumberSys) then exit;
     end
      else

    if (Sy1 <> Sy2) then exit;
   end;

  if ((Sy2 = SetSym) and not (Op in [plus, becomes, InSym])) then exit;

  if ((Sy1 in [CharSym, StringSym]) and not(Sy2 in [CharSym, StringSym])) then
   exit;

  if ((Sy2 in [CharSym, StringSym]) and not(Sy1 in [CharSym, StringSym])) then
   exit;

  if (    (Sy1 = PointSym)
      and not (Sy2 in [PointSym, NilSym])
      and not (Op in [becomes, eql, neq])) then exit;

  if (Op = becomes) then
   begin
    if ((Sy1 = CharSym) and (Sy2 = StringSym)) then exit;
    if ((Sy2 = RealSym) and (Sy1 <> RealSym)) then exit;
   end;

  if (    (Sy1 in [CharSym, StringSym])
      and not (Op in [InSym, becomes, eql, neq, lss, leq, gtr, geq, plus])) then
   exit;

  if (    (Sy1 = BoolSym)
      and not (Op in [OrSym, AndSym, XorSym, becomes, eql, neq, lss, leq,
                      gtr, geq, plus])) then exit;

  if (    (   (Sy1 = RealSym)
           or (Sy2 = RealSym))
      and (Op in [OrSym, AndSym, ShlSym, ShrSym, XorSym, DivSym, ModSym])) then
   exit;

  CompareSymbol := True;
end;


function GetCheckSymbol(sy : tSymbol) : boolean;
begin
  GetSymbol;
  GetCheckSymbol := CheckSymbol(sy);
end;


function InsertProgram : boolean;
var
  b : byte;
  tId : pIdentifier;
  tSt: pStructure;
begin
  InsertProgram := False;
  b := 0;
  while (    (ListUnits[b].ProgName <> sPasFile)
         and (ListUnits[b].ProgName <> '')
         and (b <= MaxUnits)) do inc(b);

  if (b > MaxUnits) then
   begin
    Error(200);
    exit;
   end;

  if (ListUnits[b].ProgName = sPasFile) then exit;  { already stored }

  bNowUnit := b;

  new(tSt);
  tSt^.sSymbol := ProgSym;
  tSt^.boPointer := False;

  tId := NewIdent;
  tId^.name := sPasFile;
  tId^.AssName := VarName(tId);
  tId^.UnitName := sPasFile;
  tId^.Level := 1;
  tId^.Structure := tSt;
  EnterIdGlobal(tId);

  ListUnits[b].ProgName := sPasFile;
  ListUnits[b].UnitList := nil;
  InsertProgram := True;
end;


function KonName : alfa;
begin
  KonName := 'CONST' + Dec2Hex(wKonCounter, 4);
  inc(wKonCounter);
end;


function NewIdent : pIdentifier;
var
  tId : pIdentifier;
begin
  new(tId);
  tId^.name := id;
  tId^.nlink := nil;
  tId^.Group := vars;
  tId^.Level := UnitLevel;
  tId^.boUsed := False;
  tId^.ProcVar := False;
  tId^.Structure := nil;
  NewIdent := tId;
end;


function NextLabel : Str16;
var
  s : str16;
begin
  str(LabelCounter, s);
  inc(LabelCounter);
  NextLabel := 'L' + s;
end;


function NumberSymbol(l : longint) : tSymbol;
begin
  if (l < 0) then
   begin
    if (l < -32767) then NumberSymbol := LongVal
     else NumberSymbol := IntVal;
   end
  else
   begin
    if (l < 256) then NumberSymbol := ByteVal
     else if (l < 32768) then NumberSymbol := IntVal
     else if (l < 65536) then NumberSymbol := WordVal
     else NumberSymbol:= LongVal;
   end;
end;


function Real2String(r : real) : Str80;
var s : string;
begin
  str(r:16:9, s);
  Real2String := short(s);
end;


function Search4Set(id : alfa) : pIdentifier;
var
  Found : boolean;
  Level : byte;
  tId, tId2 : pIdentifier;
begin
  Level := Top + 1;
  Found := False;

  repeat
   dec(Level);
   tId := Identifier[bNowUnit][Level];

   while ((tId <> nil) and not Found) do
    begin
     if (tId^.Structure^.sSymbol = SetSym) then
      begin
       tId2 := tId^.Structure^.tSet;

       while ((tId2 <> nil) and not Found) do
        begin
         Found := (id = tId2^.name);
         if not Found then tId2 := tId2^.nlink;
        end;
      end;

     if not Found then tId := tId^.nlink;
    end;
  until (Found or (Level = 0));

  if Found then Search4Set := tId else

{**  IMHO if a SetItemSym exists, its origin set MUST exist as well  **}
   begin
    Search4Set := nil;
    Error(201);
   end;
end;


Function SearchChainId(tId : pIdentifier; Naam : Alfa;
                      var Start : longint) : pIdentifier;
var
  Found : boolean;
begin
  Naam := short(Naam);
  Found := False;

  while ((tId <> nil) and not Found) do
   begin
    Found := (Naam = upper(short(tId^.name)));
    if not Found then
     begin
      Start := Start + tId^.Size;
      tId := tId^.nlink;
     end;
   end;

  SearchChainId := tId;
end;


function SearchId(id, UNaam : alfa; Level : byte;
                   boUsed : boolean) : pIdentifier;
var
  Naam : alfa;
  b, b2 : byte;
  plo : longint;
  tIdF, tId, tId2 : pIdentifier;
  tSt : pStructure;
  UnL : pUnitList;
  tWs1 : pWithStruc;
Label L1;

 function SubSearch(pst : alfa) : byte;
 var b : byte;
 begin
  b := 0;
  while (ListUnits[b].ProgName <> pst) do inc(b);
  SubSearch := b;
 end;

begin
  Naam := upper(short(id));
  tId := nil;
  SearchId := tId;
  b := BlockLevel;

{* First check own function/procedure *}
  while ((tId = nil) and (b > 0)) do
   begin
    tId := SearchChainId(idProcParms[b], Naam, plo);
    dec(b);
   end;

  if (tId = nil) then
   begin
    if (UNaam = '') then
     begin
      tId := SearchChainId(Identifier[bNowUnit][Level], Naam, plo);
      if (tId <> nil) then goto L1;

      tId := SearchChainId(Identifier[1][Level], Naam, plo);
      if (tId <> nil) then goto L1;

      UnL := ListUnits[bNowUnit].UnitList;
      tIdF := nil;

      while (UnL <> nil) do
       begin
        b2 := SubSearch(UnL^.UnitName);
        tId := Identifier[b2][Level];
        tId := SearchChainId(tId, Naam, plo);
        if tId <> nil then tIdF := tId;
        UnL := UnL^.NextUnit;
       end;

      tId := tIdF;
     end
    else
     tId := SearchChainId(Identifier[SubSearch(UNaam)][Level], Naam, plo);
   end;

  if (    (tId = nil) and (WithTop > 0)
      and (WithStruc[1] <> nil) and (WithStruc[1]^.ThisWith <> nil)) then
   begin
    b := 1;

    repeat
     tWs1 := WithStruc[b];
     tSt := tWs1^.ThisWith^.Structure;

     if (tSt^.sSymbol <> RecordSym) then
      begin
       repeat
        if (tSt^.sSymbol = ArraySym) then tSt := tSt^.ArStruc;

        if (tSt^.sSymbol = TypeSym) then
         repeat
          tSt := tSt^.t^.Structure;
         until (tSt^.sSymbol <> TypeSym);
       until (tSt^.sSymbol <> ArraySym);

       if (tSt^.sSymbol <> RecordSym) then
        begin
         Error(75);
         exit;
        end;
      end;

     tId2 := tSt^.Item;

     repeat
      tId := SearchChainId(tId2, Naam, plo);
      tWs1 := tWs1^.NextWith;
     until ((tId <> nil) or (tWs1 = nil));

     inc(b);
    until ((tId <> nil) or (b > WithTop));
   end;

L1:
  if (tId <> nil) and boUsed then tId^.boUsed := boUsed;
  SearchId := tId;
end;


function SearchIdAll(id, UNaam : alfa; Level : byte;
                      boUsed : boolean) : pIdentifier;
var
  p : pointer;
begin
  inc(Level);

  repeat
   dec(Level);
   p := SearchId(id, UNaam, Level, boUsed);
  until (( p <> nil) or (Level = 0));

  SearchIdAll := p;
end;


function SearchInSet(id, UNaam : alfa; tId : pIdentifier) : boolean;
var
  Found : boolean;
  tId2 : pIdentifier;
begin
  Found := False;

  if (tId^.Structure^.sSymbol = SetSym) then
   begin
    tId2 := tId^.Structure^.tSet;

    while ((tId2 <> nil) and not Found) do
     begin
      Found := (   (id = tId2^.name)
                and (   (UNaam = '')
                     or (UNaam = tId2^.UnitName)));
      if not Found then tId2 := tId2^.nlink;
     end;
   end;

  SearchInSet := Found;
end;


function SearchSet(Name : alfa; tIdS : pIdentifier) : longint;
{
 Checks if given name is part of a set.
 Input
  firts identifier of set
  name to be checked
 Output:
  Value if found
  -1 if not found
}
begin
  SearchSet := -1;
  if (tIdS = nil) then exit;

  while ((Name <> tIdS^.name) and (tIdS <> nil)) do
   tIdS := tIdS^.nlink;

  if (tIdS <> nil) then SearchSet := tIdS^.start;
end;


function Subsize(tSt : pStructure; bo : boolean) : longint;
var
  l : longint;
  tCa : pCaseStruc;
  tId2 : pIdentifier;
  tRa : pRange;
begin
  l := 0;

  if (tSt^.boPointer and bo) then Subsize := PointerSize else
   case tSt^.sSymbol of
     ArraySym  : begin
       l := 1;
       tRa := tSt^.ArRange;

       while (tRa <> nil) do
        begin
         l := l * (tRa^.lEnd - tRa^.lBegin + 1);
         tRa := tRa^.nRange;
        end;

       tSt^.ArSize := SubSize(tSt^.ArStruc, bo);
       Subsize := l * tSt^.ArSize;
      end;

     BoolSym   : Subsize := BooleanSize;
     ByteSym   : Subsize := ByteSize;
     CharSym   : Subsize := CharSize;
     FalseSym  : Subsize := BooleanSize;
     FileSym,
     TextSym   : Subsize := SizeOf(tFileDef);
     IntSym    : Subsize := IntSize;
     LabelSym  : Subsize := LabelSize;
     LongSym   : Subsize := LongSize;
     PointSym  : Subsize := PointerSize;
     RangeSym  : SubSize := tSt^.Range^.lEnd - tSt^.Range^.lBegin + 1;
     RealSym   : Subsize := RealSize;

     RecordSym : begin
       tId2 := tSt^.Item;
       while (tId2 <> nil) do
        begin
         if tId2^.Structure^.boPointer then l := l + PointerSize else
          l := l + tId2^.Size;
         tId2 := tId2^.nlink;
        end;

       if (tSt^.CaseStruc <> nil) then
        begin
         inc(l);                        { case item }
         tCa := tSt^.CaseStruc;

         while (tCa <> nil) do
          begin
           tId2 := tCa^.CsubPart;

           while (tId2 <> nil) do
            begin
             if tId2^.Structure^.boPointer then l := l + PointerSize else
              l := l + tId2^.Size;
             tId2 := tId2^.nlink;
            end;

           tCa := tCa^.NextCase;
          end;
        end;

       Subsize := l;
      end;

     SetSym    : begin
       tId2 := tSt^.tSet;

       while tId2 <> nil do
        begin
         inc(l);
         tId2 := tId2^.nlink;
        end;

       Subsize := l;
      end;

     SetOfSym  : Subsize := Subsize(tSt^.tSetOf^.Structure, bo);
     StringSym : Subsize := tSt^.StrSize;
     To_do     : Subsize := PointerSize;
     TypeSym   : Subsize := Subsize(tSt^.t^.Structure, bo);
     TrueSym   : Subsize := BooleanSize;
     WordSym   : Subsize := WordSize;
      else Subsize := 99999;
    end;
end;


function Symbol2char(s : tSymbol) : str3;
begin
  case s of
    AndSym : Symbol2Char := 'AND';
    DivSym : Symbol2Char := 'DIV';
    eql    : Symbol2Char := 'EQL';
    geq    : Symbol2Char := 'GEQ';
    gtr    : Symbol2Char := 'GTR';
    leq    : Symbol2Char := 'LEQ';
    lss    : Symbol2Char := 'LSS';
    minus  : Symbol2Char := '-';
    ModSym : Symbol2Char := 'MOD';
    neq    : Symbol2Char := 'NEQ';
    NotSym : Symbol2Char := 'NOT';
    OrSym  : Symbol2Char := 'OR';
    plus   : Symbol2Char := '+';
    ShlSym : Symbol2Char := 'SHL';
    ShrSym : Symbol2Char := 'SHR';
    slash  : Symbol2Char := '/';
    star   : Symbol2Char := '*';
    XorSym : Symbol2Char := 'XOR';
     else Symbol2Char := '???';
   end;
end;


function Symbol2Num(s : tSymbol) : byte;
begin
  case s of
    BoolSym    : Symbol2Num := 0;
    ByteSym    : Symbol2Num := 1;
    CharSym    : Symbol2Num := 2;
    IntSym     : Symbol2Num := 3;
    WordSym    : Symbol2Num := 4;
    LongSym    : Symbol2Num := 5;
    PointSym   : Symbol2Num := 6;
    RealSym    : Symbol2Num := 7;
    SetItemSym : Symbol2Num := 8;
    StringSym  : Symbol2Num := 9;
   else Symbol2Num := $FF;
  end;
end;


function Symbol2Size(s : tSymbol) : word;
begin
  case s of
    BoolSym   : Symbol2Size := Booleansize;
    ByteSym   : Symbol2Size := ByteSize;
    CharSym   : Symbol2Size := CharSize;
    IntSym    : Symbol2Size := IntSize;
    WordSym   : Symbol2Size := WordSize;
    LongSym   : Symbol2Size := LongSize;
    RealSym   : Symbol2Size := RealSize;
    StringSym : Symbol2Size := StringSize + 1;
   else Symbol2Size := 0;
  end;
end;


function Symbol2Type(s : tSymbol) : Str3;
var t : str3;
begin
  case s of
    BoolSym    : t := '00';
    ByteSym    : t := '01';
    CharSym    : t := '02';
    IntSym     : t := '03';
    WordSym    : t := '04';
    LongSym    : t := '05';
    PointSym   : t := '06';
    RealSym    : t := '07';
    SetItemSym : t := '08';
    StringSym  : t := '09';
   else t := 'FF';
  end;

  Symbol2Type := '$' + t;
end;


function tReSize(tRe : Result) : word;
var s : word;
begin
  case tRe.Symbol of
    ByteSym   : s := ByteSize;
    CharSym   : s := CharSize;
    WordSym   : s := WordSize;
    IntSym    : s := IntSize;
    LongSym   : s := LongSize;
    RealSym   : s := RealSize;
    StringSym : if tRe.Group = Konst then s := ord(tRe.s^[0]) + 1
                 else s := StringSize + 1;
   end;
  tReSize := s;
end;


function Val2Sym(s : tSymbol) : tSymbol;
begin
  case s of
    ByteVal   : Val2Sym := ByteSym;
    CharVal   : Val2Sym := CharSym;
    IntVal    : Val2Sym := IntSym;
    LongVal   : Val2Sym := LongSym;
    RangeVal  : Val2Sym := RangeSym;
    RealVal   : Val2Sym := RealSym;
    StringVal : Val2Sym := StringSym;
    WordVal   : Val2Sym := WordSym
     else Val2Sym := s;
   end;
end;


function VarName(tId : pIdentifier) : alfa;
var c : char;
begin
  if Top = 0 then c := 'V' else c := 'S';
  VarName := c + Dec2Hex(wVarCounter, 4) + short(tId^.name);
  inc(wVarCounter);
end;


{============================================================================}


procedure CalcStart(tId : pIdentifier);
{
 Calculate the relative address of the given const/variable.
}
begin
  tId^.Size := Subsize(tId^.Structure, True);
  if (BlockLevel = 0) then tId^.Start := lStart;
  if (UpdateStart and (BlockLevel = 0)) then lStart := lStart + tId^.Size;
  if (BlockLevel <> 0) then tId^.Start := lStack;
end;


procedure CalcResult(tSt, tSt2 : pStructure; Op1 : tSymbol);
{
 Calculate the result of both found terms
}
var
  r1, r2 : real;
  str1, str2 : string;

 procedure GoReal;
 begin
   if (tSt^.sSymbol = RealSym) then r1 := tSt^.r
    else r1 := tSt^.l;
   if (tSt2^.sSymbol = RealSym) then r2 := tSt2^.r
    else r2 := tSt2^.l;
   tSt^.sSymbol := RealSym;
 end;

 function TermIsReal : boolean;
 begin
   if ((tSt^.sSymbol = RealSym) or (tSt2^.sSymbol = RealSym)) then
    begin
     GoReal;
     TermIsReal := true;
    end
     else TermIsReal := False;
 end;

begin
  if (tSt^.sSymbol in [CharSym, StringSym]) then
   begin
    if (tSt^.sSymbol = CharSym) then Str1 := tSt^.c
     else Str1 := tSt^.s^;

    if (tSt2^.sSymbol = CharSym) then Str2 := tSt2^.c
     else Str2 := tSt2^.s^;

    if (Op1 = plus) then
     begin
      if (tSt^.sSymbol <> StringSym) then
       begin
        tSt^.sSymbol := StringSym;
        new(tSt^.s);
       end;

      tSt^.s^ := str1 + str2;
      tSt^.StrSize := length(tSt^.s^) + 1;
     end
    else
     begin
      tSt^.sSymbol := BoolSym;

      if (Op1 = eql) then tSt^.b := (Str1 = Str2)
       else if (Op1 = neq) then tSt^.b := (Str1 <> Str2)
       else if (Op1 = lss) then tSt^.b := (Str1 < Str2)
       else if (Op1 = leq) then tSt^.b := (Str1 <= Str2)
       else if (Op1 = gtr) then tSt^.b := (Str1 > Str2)
       else if (Op1 = geq) then tSt^.b := (Str1 >= Str2);
     end;
   end
    else

  if (Op1 = slash) then
   begin
    GoReal;
    tSt^.r := r1 / r2;
   end
    else

  if (Op1 = plus) then
   begin
    if TermIsReal then tSt^.r := r1 + r2 else
     begin
      tSt^.l := tSt^.l + tSt2^.l;
      tSt^.sSymbol := Val2Sym(NumberSymbol(tSt^.l));
     end;
   end
    else

  if (Op1 = minus) then
   begin
    if TermIsReal then tSt^.r := r1 - r2 else
     begin
      tSt^.l := tSt^.l - tSt2^.l;
      tSt^.sSymbol := Val2Sym(NumberSymbol(tSt^.l));
     end;
   end
    else

  if (Op1 = star) then
   begin
    if TermIsReal then tSt^.r := r1 * r2 else
     begin
      tSt^.l := tSt^.l * tSt2^.l;
      tSt^.sSymbol := Val2Sym(NumberSymbol(tSt^.l));
     end;
   end
    else

  if (Op1 = AndSym) then
   begin
    tSt^.l := tSt^.l and tSt2^.l;
    tSt^.sSymbol := Val2Sym(NumberSymbol(tSt^.l));
   end
    else

  if (Op1 = DivSym) then
   begin
    tSt^.l := tSt^.l div tSt2^.l;
    tSt^.sSymbol := Val2Sym(NumberSymbol(tSt^.l));
   end
    else

  if (Op1 = ModSym) then
   begin
    tSt^.l := tSt^.l mod tSt2^.l;
    tSt^.sSymbol := Val2Sym(NumberSymbol(tSt^.l));
   end
    else

  if (Op1 = OrSym) then
   begin
    tSt^.l := tSt^.l or tSt2^.l;
    tSt^.sSymbol := Val2Sym(NumberSymbol(tSt^.l));
   end
    else

  if (Op1 = XorSym) then
   begin
    tSt^.l := tSt^.l xor tSt2^.l;
    tSt^.sSymbol := Val2Sym(NumberSymbol(tSt^.l));
   end;
end;


procedure CalcTerm(tSt : pStructure);
{
 CalcTerm calculates a term for CONST to a simple type like byte, char etc.
 including the value
}
var
  i : integer;
  tId : pIdentifier;
  sy : tSymbol;
begin
  if (Symbol in [plus, minus]) then
   begin
    sy := Symbol;
    GetSymbol;
    CalcTerm(tSt);

    if not (tSt^.sSymbol in [RealSym] + NumberSys) then
     begin
      Error(21);
      tSt^.sSymbol := UnknownSym;
      exit;
     end;

    if (sy = minus) then
     begin
      if (tSt^.sSymbol = RealSym) then
       begin
        tSt^.r := -tSt^.r;
        if tSt^.rs <> nil then tSt^.rs^ := '-' + tSt^.rs^;
       end
      else tSt^.l := -tSt^.l;
     end;

    exit;
   end
    else

  if (Symbol = NotSym) then
   begin
    GetSymbol;
    CalcTerm(tSt);

    if not (tSt^.sSymbol in [BoolSym] + NumberSys) then
     begin
      Error(21);
      tSt^.sSymbol := UnknownSym;
      exit;
     end;

    if (tSt^.sSymbol = BoolSym) then tSt^.b := not tSt^.b
      else tSt^.l := not tSt^.l;

    exit;
   end
    else

  if (Symbol = lParent) then
   begin
    GetSymbol;
    Calculation(tSt, 0);

    if not CheckSymbol(rParent) then exit;
   end
    else

  if (Symbol = RangeVal) then
   begin
    sString[0] := chr(ord(sString[0]) - 2);
    val(sString, tSt^.l, i);            { no check on i needed }
    tSt^.sSymbol := LongSym;
    Symbol := RangeSym;
    exit;                               { skip GetSymbol }
   end
    else

  if (Symbol = StanFunc) then
   begin
    if (fSymbol = AbsSym) then
     begin
      GetSymbol;
      CalcTerm(tSt);

      if not (tSt^.sSymbol in [RealSym] + NumberSys) then
       begin
        Error(21);
        tSt^.sSymbol := UnknownSym;
        exit;
       end;

      tSt^.l := abs(tSt^.l);
      tSt^.r := abs(tSt^.r);
     end
      else

    if (fSymbol = ChrSym) then
     begin
      GetSymbol;
      CalcTerm(tSt);

      if not (    (tSt^.sSymbol in NumberSys)
              and (tSt^.l >= 0) and (tSt^.l <= 255)) then
       begin
        Error(21);
        tSt^.sSymbol := UnknownSym;
        exit;
       end;

      tSt^.c := chr(tSt^.l);
      tSt^.sSymbol := CharSym;
     end
      else

    if (fSymbol = HiSym) then
     begin
      GetSymbol;
      CalcTerm(tSt);

      if not (tSt^.sSymbol in NumberSys) then
       begin
        Error(21);
        tSt^.sSymbol := UnknownSym;
        exit;
       end;

      tSt^.l := hi(tSt^.l);
      tSt^.sSymbol := ByteSym;
     end
      else

    if (fSymbol = LengthSym) then
     begin
      GetSymbol;
      CalcTerm(tSt);

      if not (tSt^.sSymbol in [CharSym, StringSym]) then
       begin
        Error(21);
        tSt^.sSymbol := UnknownSym;
        exit;
       end;

      if (tSt^.sSymbol = CharSym) then tSt^.l := 1
       else tSt^.l := length(tSt^.s^);
      tSt^.sSymbol := ByteSym;
     end
      else

    if (fSymbol = LoSym) then
     begin
      GetSymbol;
      CalcTerm(tSt);

      if not (tSt^.sSymbol in NumberSys) then
       begin
        Error(21);
        tSt^.sSymbol := UnknownSym;
        exit;
       end;

      tSt^.l := lo(tSt^.l);
      tSt^.sSymbol := ByteSym;
     end
      else

    if (fSymbol = OddSym) then
     begin
      GetSymbol;
      CalcTerm(tSt);

      if not (tSt^.sSymbol in NumberSys) then
       begin
        Error(21);
        tSt^.sSymbol := UnknownSym;
        exit;
       end;

      tSt^.b := odd(tSt^.l);
      tSt^.sSymbol := BoolSym;
     end
      else

    if (fSymbol = OrdSym) then
     begin
      GetSymbol;
      CalcTerm(tSt);

      if (tSt^.sSymbol <> CharSym) then
       begin
        Error(21);
        tSt^.sSymbol := UnknownSym;
        exit;
       end;

      tSt^.l := ord(tSt^.c);
      tSt^.sSymbol := ByteSym;
     end
      else

    if (fSymbol = PredSym) then
     begin
      GetSymbol;
      CalcTerm(tSt);

      if not (tSt^.sSymbol in [CharSym] + NumberSys) then
       begin
        Error(21);
        tSt^.sSymbol := UnknownSym;
        exit;
       end;

      tSt^.l := pred(tSt^.l);
      tSt^.c := pred(tSt^.c);
     end
      else

    if (fSymbol = RoundSym) then
     begin
      GetSymbol;
      CalcTerm(tSt);

      if not (tSt^.sSymbol in [RealSym] + NumberSys) then
       begin
        Error(21);
        tSt^.sSymbol := UnknownSym;
        exit;
       end;

      if (tSt^.sSymbol = RealSym) then tSt^.l := round(tSt^.r);
      tSt^.sSymbol := LongSym;
     end
      else

    if (fSymbol = SuccSym) then
     begin
      GetSymbol;
      CalcTerm(tSt);

      if not (tSt^.sSymbol in [CharSym] + NumberSys) then
       begin
        Error(21);
        tSt^.sSymbol := UnknownSym;
        exit;
       end;

      tSt^.l := Succ(tSt^.l);
      tSt^.c := Succ(tSt^.c);
     end
      else

    if (fSymbol = SwapSym) then
     begin
      GetSymbol;
      CalcTerm(tSt);

      if not (tSt^.sSymbol in [ByteSym, IntSym, WordSym]) then
       begin
        Error(21);
        tSt^.sSymbol := UnknownSym;
        exit;
       end;

      if (tSt^.sSymbol = ByteSym) then tSt^.l := 0
       else tSt^.l := swap(tSt^.l);
     end
      else

    if (fSymbol = TruncSym) then
     begin
      GetSymbol;
      CalcTerm(tSt);

      if not (tSt^.sSymbol in [RealSym] + NumberSys) then
       begin
        Error(21);
        tSt^.sSymbol := UnknownSym;
        exit;
       end;

      if (tSt^.sSymbol = RealSym) then tSt^.l := trunc(tSt^.r);
      tSt^.sSymbol := LongSym;
     end;

    exit;                               { skip GetSymbol }
   end
    else

  if (Symbol = Ident) then
   begin
    tId := SearchIdAll(Id, '', Top, False);
    if (tId = nil) then begin Error(3); exit; end;

    CopyStructure(tst, tId^.Structure);
   end
  else
   begin
    tSt^.sSymbol := Val2Sym(Symbol);


{**  Simple type is found, so what is the actual value?  **}
    if (tSt^.sSymbol in NumberSys) then tSt^.l := lFound else
    if (tSt^.sSymbol = CharSym) then tSt^.c := sString[1] else
    if (tSt^.sSymbol = RealSym) then
     begin
      tSt^.r := rFound;
      new(tSt^.rs);
      tSt^.rs^ := sString;
     end
      else

    if (tSt^.sSymbol = StringSym) then
     begin
      new(tSt^.s);
      tSt^.StrSize := length(sString);
      tSt^.s^ := sString;
     end
      else

    if (tSt^.sSymbol in [FalseSym, TrueSym]) then
     begin
      tSt^.sSymbol := BoolSym;
      tSt^.b := (Symbol = TrueSym);
     end
    else
     begin
      Error(21);
      exit;
     end;
  end;

  GetSymbol;
end;


procedure Calculation(tSt : pStructure; level : byte);
{
 Calculate arithmic constructions like "a+b-c+d*e div f mod g"
}
var
  tOrder : byte;
  tSt2 : pStructure;
  Op1 : tSymbol;                        { found operators }

begin
  CalcTerm(tSt);                        { first term }

  while not (Symbol in [comma, semicolon, rParent, rBrack, RangeSym]) do
   begin
    if tSt^.sSymbol = RealSym then
     if tSt^.rs <> nil then
      begin
       dispose(tSt^.rs);
       tSt^.rs := nil;
      end;

    if not (Symbol in [eql, neq, lss, leq, gtr, geq, plus, minus, slash,
           star, DivSym, ModSym, OrSym, AndSym, ShlSym, ShrSym, XorSym]) then
     begin
      Error(42);
      exit;
     end;

{**  What is the level of the operator?  **}
    case Symbol of
      eql, neq, lss, leq, gtr, geq : tOrder := 1;
      plus, minus, OrSym, XorSym   : tOrder := 2;
      slash, star, DivSym, ModSym, AndSym, ShlSym, ShrSym : tOrder := 3;
     end;

    if (tOrder <= Level) then exit;     { same or less the previous -> exit }

    Op1 := Symbol;
    new(tSt2);
    GetSymbol;
    tSt^.boPointer := False;
    Calculation(tSt2, tOrder);          { get second term }

{**  Check if second term matches first one using the found operator  **}
    if not CompareSymbol(tSt^.sSymbol, tSt2^.sSymbol, Op1) then
     begin
      tSt^.sSymbol := UnknownSym;
      Error(26);
      DisposeStructure(tSt2);
      exit;
     end;

    CalcResult(tSt, tSt2, Op1);
    DisposeStructure(tSt2);
   end;  {while}                {!! dirty way: use goto !!}
end;


procedure ClearTre(var tRe : Result);
begin
  tRe.Name := '';
  tRe.OrgName := '';
  tRe.Symbol := UnknownSym;
  tRe.Group := Unknown;
  tRe.boDirect := True;
  tRe.b := False;             {##  for debugging }
  tRe.l := 0;                 {##  for debugging }
  tRe.r := 0;                 {##  for debugging }
  tRe.s := nil;
  tRe.Range := nil;
  tRe.Ident := nil;
end;


procedure Compiler1Directive;
(*-------------------------------------------------------------------------
Handles the original Turbo Pascal directives {$x}
-------------------------------------------------------------------------*)
begin
  repeat GetChar until ch = '}';
  ch := ' ';
  Symbol := SemiColon;
  GetChar;
  GetSymbol;
end;


procedure Compiler2Directive;
(*-------------------------------------------------------------------------
Handles the directives {&x} for this 6502 compiler.
&COM=<computer>         C128
                        C64
                        CBM40  PET2001, 30xx, 40xx
                        CBM6   610
                        CBM7   700, 710, 720
                        CBM80  8032, 8096, 8296
                        VIC20
&CPU=<processor>        6502
                        6510
                        65816
                        65C02
                        65GZ032
&HEA=<address>    Where to store the heap
                        0      before the program
                        1      after the program/variables
                        $xxxx  on the given address
&ORG=<address>    Start of adress of the program
                        0      use '1 SYS 2063' and save as BASIC program
                        $xxxx  save as binary
&STK=<address>    Where to store the stack
                        0      at the end of the RAM
                        $xxxx  on the given address
&VAR=<address>    Where to store variables
                        0      before the program
                        1      after the program
                        $xxxx  on the given address
-------------------------------------------------------------------------*)
var
  i : integer;
  c, s : str16;
begin
  s := '';

  repeat
   GetChar;
   s := s + upcase(ch);
  until ch = '}';

  if s[4] <> '=' then
   begin
    Error(215);
    exit;
   end;

  c := copy(s, 1, 3);
  delete(s, 1, 4);
  s[0] := pred(s[0]);
  s := short(s);

  if c = 'COM' then
   begin
    sMachine := s + '.P65';

    if ((s = 'C64') or (s = 'C128') or (s = 'VIC20')) then
     lBOrgAddress := $0801 else lBOrgAddress := $0401;
   end
    else
  if c = 'CPU' then
   begin
    if ((s = '6502') or (s = '6510') or (s = '65816') or (s = '65C02') or
        (s = '65GZ032')) then
     sCpu := s
    else
     begin
      Error(216);
      exit;
     end;
   end
    else
  if c = 'ORG' then
   begin
    if S = 'BASIC' then lStartAddr := 0 else
     begin
      val(S, lStartAddr, i);

      if i <> 0 then
       begin
        Error(216);
        exit;
       end;
     end;
   end;

  ch := ' ';
  Symbol := SemiColon;
  GetChar;
  GetSymbol;
end;


procedure CopyASMheader(stFilename : str80);
var
  ft : text;
  pst : str80;
begin
  assign(ft, stFilename + '.P65');
  reset(ft);

  repeat
   readln(ft, pst);
   Wr2ASMheaderLn(pst);
  until EOF(ft);

  close(ft);
end;


procedure CopyStructure(tSt, tSt2 : pStructure);
begin
  tSt^ := tSt2^;

  if (tSt^.sSymbol = RealSym) then
   if tSt2^.rs <> nil then
    begin
     new(tSt^.rs);
     tSt^.rs^ := tSt2^.rs^;
    end;

  if (tSt^.sSymbol = StringSym) then
   begin
    new(tSt^.s);
    tSt^.s^ := tSt2^.s^;
   end;
end;


procedure CreateTextStruc(tSt : pStructure);
begin
  tSt^.sSymbol := TextSym;
  new(tSt^.f);
  tSt^.f^.Filename := '';
  tSt^.f^.Attribute := $FF;
  tSt^.f^.fType := TextSym;
  tSt^.f^.fTypeId := nil;
end;


procedure DisposeIdentifier(var tIds : pIdentifier);
var
  tId, tId2 : pIdentifier;
  tSt1, tSt2 : pStructure;
begin
  if (tIds = nil) then exit;
  tId := tIds;
  tSt1 := nil;
  tSt2 := nil;

  repeat
{write(tID^.Name, '   ');}   {####}
   tId2 := tId^.nlink;

   if (tId^.Structure) <> nil then
    begin
     tSt2 := tId^.Structure;
     if (tSt1 <> tSt2) then DisposeStructure(tSt1);
     tSt1 := tSt2;
    end;

   dispose(tId);
   tId := tId2;
  until tId = nil;

  DisposeStructure(tSt1);
  tIds := nil;
end;


procedure DisposeRange(var tRa1 : pRange);
var
  tRa2, tRa3 : pRange;
begin
  if (tRa1 = nil) then exit;
  tRa2 := tRa1;

  repeat
   tRa3 := tRa2^.nRange;
   dispose(tRa2);
   tRa2 := tRa3;
  until tRa2 = nil;

  tRa1 := nil;
end;


procedure DisposeStructure(var tSts : pStructure);
var
  tSt : pStructure;
  tCa1, tCa2 : pCaseStruc;
begin
  if (tSts = nil) then exit;
  tSt := tSts;

  case tSt^.sSymbol of
    ArraySym   : begin
      DisposeRange(tSt^.ArRange);
      tSt^.ArRange := nil;
      DisposeStructure(tSt^.ArStruc);
      if (tSt^.DataSize > 0) then FreeMem(tSt^.DataPtr, tSt^.DataSize);
     end;
    TextSym,
    FileSym : dispose(tSt^.f);
    FuncSym     : begin
      if ( tSt^.OrgDeclaration <> nil) then dispose(tSt^.OrgDeclaration);
      DisposeIdentifier(tSt^.Parameters);
      Dispose(tSt^.Functie);
     end;
    ProcSym : begin
      if ( tSt^.OrgDeclaration <> nil) then dispose(tSt^.OrgDeclaration);
      DisposeIdentifier(tSt^.Parameters);
     end;
    RangeSym : DisposeRange(tSt^.Range);
    RecordSym : begin
      DisposeIdentifier(tSt^.Item);
      DisposeIdentifier(tSt^.CaseItem);

      if (tSt^.CaseStruc <> nil) then
       begin
        tCa1 := tSt^.CaseStruc;

        while (tCa1 <> nil) do
         begin
          tCa2 := tCa1^.NextCase;
          if (tCa1^.CsubType = StringVal) then dispose(tCa1^.s);
          dispose(tCa1);
          tCa1 := tCa2;
         end;

        tSt^.CaseStruc := nil;
       end;

      DisposeIdentifier(tSt^.CaseMembers);
     end;
    RealSym : if (tSt^.rs <> nil) then
     begin
      dispose(tSt^.rs);
      tSt^.rs := nil;
     end;
    StringSym : if (tSt^.s <> nil) then
     begin
      dispose(tSt^.s);
      tSt^.s := nil;
     end;
   end;

  dispose(tSts);
  tSts := nil;
end;


procedure EnterIdGlobal(tIdent: pIdentifier);
{
 enter id pointed at by tIdent into the name-table
}
var
  tId, tId1 : pIdentifier;
begin
  EnterIdLocal(tIdent,  Identifier[bNowUnit][top]);
end;


procedure EnterIdLocal(tIdent : pIdentifier; var tIdD : pIdentifier);
{
 enter id pointed at by tIdent into the name-table
}
var
  tId, tId1 : pIdentifier;
begin
  tId := tIdD;

  if tId = nil then tIdD := tIdent else
   begin
    repeat
     tId1 := tId;
     if (tId^.name = tIdent^.name) then Error(4);      {name already exists}
     tId := tId^.nlink;
    until tId = nil;

    tId1^.nlink := tIdent
   end;
end;


procedure Error(ErrNr : integer);
var by : byte;
begin
  if boReported then exit;

wError := ErrNr;

  writeln;
  writeln('Error at line: ', Linecount, '     errorcode: ', ErrNr);
  writeln(sRegel);
  writeln;
  for by := 1 to chCounter-4 do write(' ');
  writeln('^');                         { point to error }

  boReported := True;
  Symbol := UnknownSym;
end {error} ;


procedure GetChar;
begin
  if EndOfLine then chCounter := 0;

  if not eof(fInput) then
   begin
    read(fInput,ch);
    EndOfLine := eoln(fInput);
    inc(chCounter);
   end
  else
   begin
    if PrevSymbol = EndSym then ch := ' ' else
     Error(10);
   end;

  if (sRegel[length(sRegel)] in [#10, #13]) then sRegel := '';
  sRegel := sRegel + ch;
  if (boASMcode or not (ch in [#10, #13])) then PasCode := PasCode + ch;
  if (ch = #10) then inc(LineCount);
end;


procedure GetFile(tSt : pStructure);
var
  tId : pIdentifier;
label L1;
begin
  new(tSt^.f);
  tSt^.f^.Filename := '';
  tSt^.f^.Attribute := $FF;
  tSt^.f^.fType := FileSym;
  tSt^.f^.fTypeId := nil;

  if (Symbol = semicolon) then exit;    { just  xxx : File; }

  if not CheckSymbol(OfSym) then exit;

  GetSymbol;
  tSt^.f^.fType := Symbol;

  if (Symbol in [ByteSym, CharSym, IntSym, LongSym, RealSym, StringSym]) then
   tSt^.f^.fType := Symbol
    else

  if (Symbol = ident) then
   begin
    tId := SearchIdAll(id, '', Top, True);

    if (   (tId = nil)
        or (tId^.Group <> Types)
        or (tId^.Structure^.sSymbol = TypeSym)) then goto L1;

    tSt^.f^.fType := TypeSym;
    tSt^.f^.fTypeId := tId;
   end
  else
   begin                                { other types not accepted }
L1:
    tSt^.f^.fType := UnknownSym;
    Error(21);
    exit;
   end;

  GetSymbol;
end;


procedure GetSymbol;
{--------------------------------------------------------------------------
Reads one or more characters from the text and translates it to a symbol
--------------------------------------------------------------------------}
var
  boDot, boE, ErrorFound : boolean;
  b, k : byte;
  c : char;
  i : integer;
  pst : str16;
  s : tSymbol;
label PL1, PL2, PL3, Pl4;

  function GetNumber(boHex : boolean) : boolean;
  begin
    repeat
     if (ord(sString[0]) = MaxDigits) then
      begin
       GetNumber := True;
       exit;
      end;

     GetNumber := False;
     sString := sString + upcase(ch);
     GetChar;
    until not(   (ch in ['0'..'9'])
              or (boHex and (upcase(ch) in ['A'..'F'])));
  end;

begin
  while (ch in [#9, #10, #13, ' ']) do GetChar; { ignore white space }
  PrevSymbol := Symbol;
  Symbol := UnknownSym;
  sString := '';

  if ch in ['a'..'z', 'A'..'Z', '_'] then       { found text/identifier }
   begin
    k:= 0;
    id := '';

    repeat
     if (k < MaxLngIdent) then
      begin
       inc(k);
       id := id + ch;
      end;

     GetChar;
    until not (ch in ['a'..'z', 'A'..'Z', '0'..'9', '_']);

    while id[0] < chr(MaxLngIdent) do id := id + ' ';
    Symbol := ident;

    for s := AndSym to XorSym do
     if Upper(Keyword[s]) = Upper(id) then
      begin
       Symbol := s;
       break;
      end;

    if Symbol = ident then
     for s := AbsSym to TruncSym do
      if Upper(aFunction[s]) = Upper(id) then
       begin
        Symbol := StanFunc;
        fSymbol := s;
        break;
       end;

    if (Symbol = ident) then id := Upper(id);
   end
    else

  if ch in ['0'..'9'] then                      { found a (real) number }
   begin
    boDot := False;
    boE := false;

    if GetNumber(False) then goto PL1;

    if ch = '.' then                            { fraction detected }
     begin
      boDot := True;

      if (ord(sString[0]) = MaxDigits) then
       begin
PL1:    Error(8);
        exit;
       end;

      sString := sString + ch;
      GetChar;

      if (ch = '.') then                        { second dot detected }
       begin
        sString := sString + ch;
        Symbol:= RangeVal;
        GetChar;
        exit;
       end;

      if (  (ord(sString[0]) = MaxDigits)
          or not(ch in ['0'..'9'])) then goto PL1;

      if GetNumber(False) then goto PL1;
     end;

    if Upcase(ch) = 'E' then                    { power of 10 }
     begin
      if boE then goto PL1;                     { second "e" detected }
      boE := True;

      if (ord(sString[0]) = MaxDigits) then goto PL1;

      sString := sString + ch;
      GetChar;

      if (  (ord(sString[0]) = MaxDigits)
          or  not(ch in ['+', '-', '0'..'9'])) then goto PL1;

      if ch in ['+', '-'] then
       begin
        sString := sString + ch;
        GetChar;

        if (  (ord(sString[0]) = MaxDigits)
            or not(ch in ['0'..'9'])) then goto PL1;
       end;

      if GetNumber(False) then goto PL1;
     end;

    val(sString, rFound, i);
    {### error 205, floating point overflow }

    if (i > 0) then goto PL1;
                                                { real number }
    if (   boDot or boE                         { outside longint range }
        or (rfound > $7FFFFFFF)
        or (rfound < -$7FFFFFFF)) then Symbol:= RealVal else
     begin
PL2:  val(sString, lFound, i);
      if i > 0 then goto PL1;
      Symbol := NumberSymbol(lFound);
     end;
   end
    else

  if ch = '$' then                              { hexidecimal }
   begin
    if GetNumber(True) then goto PL1
     else goto PL2
   end
    else

  if ch = ':' then
   begin
    GetChar;
    if ch = '=' then
     begin
      Symbol:= becomes;
      GetChar;
     end
      else Symbol:= colon;
   end
    else

  if ch = '>' then
   begin
    GetChar;

    if ch = '=' then
     begin
      Symbol:= geq;
      GetChar;
     end
      else Symbol:= gtr
   end
    else

  if ch = '<' then
   begin
    GetChar;

    if ch = '=' then
     begin
      Symbol:= leq;
      GetChar;
     end
      else
     if ch='>' then
      begin
       Symbol:= neq;
       GetChar;
      end
       else Symbol:= lss;
   end
    else

  if ch = '.' then
   begin
    GetChar;

    if ch = '.' then
     begin
      Symbol:= RangeSym;
      GetChar;
     end
      else Symbol:= period;
   end
    else

  if ch = '''' then                             { start of char or string }
   begin
    sString := '';
    Symbol := CharVal;

PL3:
    ErrorFound := EndOfLine;

    if not ErrorFound then
     repeat
      repeat
       GetChar;
       sString := sString + ch;
       ErrorFound := (ord(sString[0]) = 255) or (EndOfLine and (ch <> ''''));
      until (ch = '''') or ErrorFound;
{
 At this point even number of quotes, thus no more text can be expected
}

      if not ErrorFound then GetChar;
     until (ch <> '''') or ErrorFound;

    sString[0] := pred(sString[0]);             { delete last quote }
    if (ord(sString[0]) <> 1) then Symbol := StringVal;
    if (ch = '#') then goto PL4;

    if ErrorFound then
     begin
      Error(8);
      exit;
     end;
   end
    else

  if ch='(' then
   begin
    GetChar;

    if (ch = '*') then                { disregard comment between (* and *) }
     begin
      boComment := False;
      b := length(PasCode) - 2;
      ch := ' ';                                { erase ch }

      repeat
       c := ch;
       GetChar;
      until ((c = '*') and (ch = ')'));

      boComment := True;
      PasCode[0] := chr(b);
      GetChar;
      GetSymbol;
     end
      else Symbol:= lParent;
   end
    else

  if ch='{' then
   begin
    b := length(PasCode) - 1;
    GetChar;

    if ch = '$' then Symbol := CompDir1Sym else
    if ch = '&' then Symbol := CompDir2Sym else
     begin
      boComment := False;
      repeat GetChar until ch='}';
      boComment := True;
      PasCode[0] := chr(b);
      GetChar;
      Symbol := PrevSymbol;
      GetSymbol;
     end;
   end
    else

  if ch='#' then
   begin
    Symbol := CharVal;
    sString := '';

PL4:
    pst := '';
    ch := '0';

    repeat
     pst := pst + ch;
     GetChar;
    until not(ch in ['0'..'9']);

    val(pst, lFound, i);
    if ((i > 0) or (lFound > 255)) then
     begin
      Error(8);
      exit;
     end;

    sString := sString + chr(lFound);
    if (ch = '''') then goto PL3;
    if (ch = '#') then goto PL4;
    if (ord(sString[0]) <> 1) then Symbol := StringVal;
   end

    else if ch='/' then begin GetChar; Symbol:= slash end
    else if ch='@' then begin GetChar; Symbol:= AtSym end
    else if ch='^' then begin GetChar; Symbol:= dakje end
    else if ch='+' then begin GetChar; Symbol:= plus end
    else if ch='-' then begin GetChar; Symbol:= minus end
    else if ch='*' then begin GetChar; Symbol:= star end
    else if ch=')' then begin GetChar; Symbol:= rParent end
    else if ch='[' then begin GetChar; Symbol:= lbrack end
    else if ch=']' then begin GetChar; Symbol:= rbrack end
    else if ch='=' then begin GetChar; Symbol:= eql end
    else if ch=',' then begin GetChar; Symbol:= comma end
    else if ch=';' then begin GetChar; Symbol:= semicolon end
    else error(300)
end;


Procedure HandleAbsolute;
var
  tId : pIdentifier;
begin
  boAbsolute := True;
  reAbsolute.AbsN := '';

  GetSymbol;

  if (Symbol = Ident) then
   begin
    tId := SearchIdAll(Id, '', Top, True);
    if (tId = nil) then begin Error(3); exit; end;

    if not (tId^.Group in [Konst, Vars]) then
     begin Error(21); exit; end;

    reAbsolute.AbsN := tId^.AssName;
    GetSymbol;
    exit;
   end;

  if not (Symbol in [ByteVal, WordVal, LongVal]) then
   begin Error(21); exit; end;

  reAbsolute.AbsH := lFound;
  GetSymbol;

  if (Symbol = Colon) then
   begin
    GetSymbol;

    if not (Symbol in [ByteVal, WordVal, LongVal]) then
     begin Error(21); exit; end;

    reAbsolute.AbsL := lFound;
    GetSymbol;
   end
  else
   begin
    reAbsolute.AbsL :=  reAbsolute.AbsH;
    reAbsolute.AbsH := 0;
   end;
end;


procedure InsertUnit(PName : alfa);
var
  b : byte;
  UnL : pUnitList;
begin
  b := 0;
  while (ListUnits[b].ProgName <> PName) do inc(b);
  UnL := ListUnits[b].UnitList;

  if UnL = nil then
   begin
    new(ListUnits[b].UnitList);
    ListUnits[b].UnitList^.UnitName := sPasFile;
    ListUnits[b].UnitList^.NextUnit := nil;
    exit;
   end;

  while ((UnL^.UnitName <> sPasFile) and (UnL^.NextUnit <> nil)) do
   UnL := UnL^.NextUnit;

  if (UnL^.UnitName = sPasFile) then    { unit used twice }
   begin
    Error(202);
    exit;
   end;

  new(UnL^.NextUnit);
  UnL^.NextUnit^.UnitName := sPasFile;
  UnL^.NextUnit^.NextUnit := nil;
end;


procedure SkipDirective;
begin
  repeat GetChar until ch = '}';
  ch := ' ';
  Symbol := SemiColon;
  GetChar;
  GetSymbol;
end;


procedure Struc2Result(var tRe1 : Result; tSt : pStructure; boPnt : boolean);
label L1;
begin
  tRe1.l := Subsize(tSt, False);


  if (boPnt and tSt^.boPointer) then tRe1.Symbol := PointSym
   else

  if (tSt^.sSymbol = TypeSym) then
   begin
L1:
    repeat
     tRe1.Symbol := tSt^.t^.Structure^.sSymbol;

     if (boPnt and tSt^.t^.Structure^.boPointer) then
      tRe1.Symbol := PointSym;

     if (tRe1.Symbol = ArraySym) then
      begin
       tRe1.Symbol := tSt^.t^.Structure^.ArStruc^.sSymbol;

       if tRe1.Symbol = TypeSym then
        tRe1.Symbol := tSt^.t^.Structure^.ArStruc^.t^.Structure^.sSymbol;
      end;

     if (tRe1.Symbol = StringSym) then
      tRe1.StrSize := tSt^.t^.Structure^.StrSize;

     tSt := tSt^.t^.Structure;
    until (tRe1.Symbol <> TypeSym);

    if (tRe1.Symbol = SetSym) then tRe1.Symbol := SetItemSym;
   end
  else
   begin
    tRe1.Symbol := tSt^.sSymbol;

{
    if (tRe1.Symbol = ArraySym) then
     begin
      tRe1.Symbol := tSt^.ArStruc^.sSymbol;

      if (tRe1.symbol = TypeSym) then
       begin
        tSt := tSt^.ArStruc;
        goto L1;
       end;
     end;
}

    if (tRe1.Symbol = StringSym) then
     tRe1.StrSize := tSt^.StrSize;
   end;

  if (tRe1.Group = konst) then
   begin
    case tRe1.Symbol of
      BoolSym : tRe1.b := Tst^.b;
      CharSym : tRe1.c := Tst^.c;
      RealSym : tRe1.r := Tst^.r;
      SetItemSym : tRe1.l := Tst^.si;
      StringSym : begin
        if (tRe1.s = nil) then new(tRe1.s);
        tRe1.s^ := Tst^.s^;
       end;
      else tRe1.l := Tst^.l;
     end;
   end;
end;


procedure Wr2ASMbodyLn(St : string);
begin
  St := ASMline(St);
  Writeln(ASMbody, St);
end;


procedure Wr2ASMheader(St : string);
begin
  if (   (pos('.EQ', upper(St)) > 0)
      or (pos('.BY', upper(St)) > 0)
      or (pos('.WO', upper(St)) > 0)
      or (pos('.DW', upper(St)) > 0)
      or (pos('.TX', upper(St)) > 0)) then St := ASMline(St);
  Write(ASMheader, St);
end;


procedure Wr2ASMheaderLn(St : string);
begin
  if (   (pos('.EQ', upper(St)) > 0)
      or (pos('.BY', upper(St)) > 0)
      or (pos('.WO', upper(St)) > 0)
      or (pos('.FW', upper(St)) > 0)
      or (pos('.FB', upper(St)) > 0)
      or (pos('.DW', upper(St)) > 0)
      or (pos('.TP', upper(St)) > 0)
      or (pos('.TX', upper(St)) > 0)) then St := ASMline(St);
  Writeln(ASMheader, St);
end;


procedure Wr2ASMmidLn(St : string);
begin
  St := ASMline(St);
  Writeln(ASMmid, St);
end;


procedure WriteConst;
var
  b, b2 : byte;
  i : integer;
  l : longint;
  p : pArFFFE;
  tId : pIdentifier;
  tSt : pStructure;
  tRe1 : Result;
  pst, pst2 : String;
  w, wa, ws : word;
begin
  b2 := 0;
  ClearTre(tRe1);

  while ListUnits[b2].ProgName <> '' do
   begin
    tId := Identifier[b2][0];

    while (tId <> nil) do
     begin
      if (    (tId^.Group = Konst) and tId^.boUsed
          and not (tId^.Structure^.sSymbol in [FuncSym, ProcSym, ProgSym,
                    SetItemSym])
         ) then
       begin
        pst := tId^.AssName;
        tSt := tId^.Structure;
        tRe1.Group := tId^.Group;
        Struc2Result(tRe1, tSt, False);
        tRe1.name := tId^.AssName;
        tRe1.OrgName := tId^.Name;
        tRe1.Ident := tId;

        case tRe1.Symbol of
         ArraySym : begin
           p := tSt^.DataPtr;
           wa := tSt^.ArRange^.lEnd - tSt^.ArRange^.lBegin + 1;
           ws := tSt^.ArSize;

           case Array2Symbol(tRe1.Ident^.Structure) of
            CharSym : begin
              Wr2ASMheaderLn(' .by $' + Dec2Hex(tSt^.Datasize, 2));
              pst := pst + ' .tx ''';

              for b := 0 to wa - 1 do
               pst := pst + chr(p^[b]);

              pst := pst + '''';
             end;
            ByteSym : begin
              pst := pst + ' .by $' + Dec2Hex(p^[0], 2);
              b := 1;

              while b < wa do
               begin
                pst := pst + ', $' + Dec2Hex(p^[b], 2);
                inc(b);

                if (b mod 10 = 0) and (b < wa) then
                 begin
                  Wr2ASMheaderLn(pst);
                  pst := ' .by $' + Dec2Hex(p^[b], 2);
                  inc(b)
                 end;
               end;
             end;
            IntSym : begin
              i := p^[0] + 256*p^[1];
              str(i, pst2);
              pst := pst + ' .wo ' + pst2;
              b := 1;

              while b < wa do
               begin
                i := p^[2 * b] + 256*p^[2 * b + 1];
                str(i, pst2);
                pst := pst + ', ' + pst2;
                inc(b);

                if (b mod 8 = 0) and (b < wa) then
                 begin
                  Wr2ASMheaderLn(pst);
                  i := p^[2 * b] + 256*p^[2 * b + 1];
                  str(i, pst2);
                  pst := ' .wo ' + pst2;
                  inc(b)
                 end;
               end;
             end;
            LongSym : begin
              l := p^[0] + 256*p^[1] + 65536*p^[2] + 16777216*p^[3];
              str(l, pst2);
              pst := pst + ' .dw ' + pst2;
              b := 1;

              while b < wa do
               begin
                l := p^[4*b] + 256*p^[4*b + 1] + 65536*p^[4*b + 2]
                      + 16777216*p^[4*b + 3];
                str(l, pst2);
                pst := pst + ', ' + pst2;
                inc(b);

                if (b mod 8 = 0) and (b < wa) then
                 begin
                  Wr2ASMheaderLn(pst);
                  l := p^[4*b] + 256*p^[4*b + 1] + 65536*p^[4*b + 2]
                        + 16777216*p^[4*b + 3];
                  str(l, pst2);
                  pst := ' .dw ' + pst2;
                  inc(b)
                 end;
               end;
             end;
            StringSym : begin
              i := p^[0];
              pst := pst + ' .by $' + Dec2Hex(i, 2);
              Wr2ASMheaderLn(pst);
              pst2 := '';
              for l := 1 to i do pst2 := pst2 + chr(p^[l]);

              if length(pst2) > 40 then
               begin
                pst := ' .tx ''' + copy(pst2, 1, 40) + '''';
                Wr2ASMheaderLn(pst);
                delete(pst2, 1, 40);

                while length(pst2) > 40 do
                 begin
                  pst := ' .tx ''' + copy(pst2, 1, 40) + '''';
                  Wr2ASMheaderLn(pst);
                  delete(pst2, 1, 40);
                 end;

                pst := ' .tx ''' + pst2 + '''';
               end
              else pst := ' .tx ''' + pst2 + '''';

              if i < ws-1 then
               begin
                Wr2ASMheaderLn(pst);
                pst := ' .fb 0, $' + Dec2Hex(ws - 1 - i, 2);
               end;

              for b := 1 to wa - 1 do
               begin
                Wr2ASMheaderLn(pst);
                i := p^[b*ws];
                Wr2ASMheaderLn(' .by $' + Dec2Hex(i, 2));
                pst2 := '';
                for l := 1 to i do pst2 := pst2 + chr(p^[b*ws + l]);

                while length(pst2) > 40 do
                 begin
                  pst := ' .tx ''' + copy(pst2, 1, 40) + '''';
                  Wr2ASMheaderLn(pst);
                  delete(pst2, 1, 40);
                 end;

                pst := ' .tx ''' + pst2 + '''';

                if i < ws-1 then
                 begin
                  Wr2ASMheaderLn(pst);
                  pst := ' .fb 0, $' + Dec2Hex(ws - 1 - i, 2);
                 end;
               end;
             end;
            WordSym : begin
              w := p^[0] + 256*p^[1];
              pst := pst + ' .wo $' + Dec2Hex(w, 4);
              b := 1;

              while b < wa do
               begin
                w := p^[2 * b] + 256*p^[2 * b + 1];
                pst := pst + ', $' + Dec2Hex(w, 4);
                inc(b);

                if (b mod 8 = 0) and (b < wa) then
                 begin
                  Wr2ASMheaderLn(pst);
                  w := p^[2 * b] + 256*p^[2 * b + 1];
                  pst := ' .wo $' + Dec2Hex(w, 4);
                  inc(b)
                 end;
               end;
             end;
           end;
          end;
         BoolSym :
          if tRe1.b then pst := pst + ' .by $FF'
           else pst := pst + ' .by $00';
         CharSym :
          if (tRe1.c = '''') then pst := pst + ' .tx "' + tRe1.c + '"' else
           if ord(tRe1.c) in [32..127] then
             pst := pst + ' .tx ''' + tRe1.c + ''''
            else pst := pst + ' .by $' + Dec2Hex(ord(tRe1.c), 2);
         SetItemSym,
         ByteSym : pst := pst + ' .by $' + Dec2Hex(tRe1.l, 2);
         IntSym,
         WordSym : pst := pst + ' .wo $' + Dec2Hex(tRe1.l, 4);
         PointSym,
         LongSym : pst := pst + ' .dw $' + Dec2Hex(tRe1.l, 8);
         RealSym : begin

{ Optimised for multiple runs }
           Wr2ASMmidLn(' macString2Real ' + pst);
           Wr2ASMheaderLn(' .by 0');

           if tSt^.rs = nil then pst2 := Real2String(tRe1.r)
            else pst2 := tSt^.rs^;

           i := length(pst2);
           pst := pst + ' .tx "' + pst2 + '"';
           Wr2ASMheaderLn(pst);

           if i < 6 then
            begin
             str(6-i, pst);
             pst := ' .fb 0, ' + pst;
            end
           else pst := ' .by 0';
          end;
         StringSym : begin
           if pos('''', tRe1.s^) = 0 then
            begin
             Wr2ASMheaderLn(pst + ' .by $' + Dec2Hex(length(tRe1.s^), 2));
             if tRe1.s^ = '' then tRe1.s^ := 'R';
             pst := ' .tx ''' + tRe1.s^ + '''';
            end
           else
            pst := pst + ' .tp $' + Dec2Hex(tRe1.StrSize, 4)
                                + ', ' + tRe1.s^;
           dispose(tRe1.s);
           tRe1.s := nil;
          end;
        end;

        Wr2ASMheaderLn(pst);
       end;

      tId := tId^.nlink;
     end;

    inc(b2);
   end;
end;


procedure WriteVars;
var
  b : byte;
  tId : pIdentifier;
  pst : Str80;
begin
  b := 0;

  while ListUnits[b].ProgName <> '' do
   begin
    tId := Identifier[b][0];

    while (tId <> nil) do
     begin
      if (    (tId^.Group = Vars) and tId^.boUsed
          and not (tId^.Structure^.sSymbol in [FuncSym, ProcSym, ProgSym,
                    SetItemSym])
         ) then
       begin
        pst := tId^.AssName + ' .fb $00, $' + Dec2Hex(tId^.Size, 4);

        if (PlaceVar.bType = 1) then Wr2ASMbodyLn(pst)
         else Wr2ASMheaderLn(pst);
       end;

      tId := tId^.nlink;
     end;

    inc(b);
   end;

  if (PlaceVar.bType = 1) then Wr2ASMbodyLn('')
   else Wr2ASMheaderLn('');
end;


begin
  boStored00 := False;
  boStoredFF := False;
end.



